package com.yonge.netty.server.service;

import io.netty.channel.Channel;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import javax.sound.sampled.AudioFormat;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import com.yonge.audio.analysis.AudioFloatConverter;
import com.yonge.audio.utils.ArrayUtil;
import com.yonge.nettty.dto.UserChannelContext;
import com.yonge.nettty.dto.WebSocketResponse;
import com.yonge.nettty.entity.MusicXmlBasicInfo;
import com.yonge.netty.server.handler.NettyChannelManager;
import com.yonge.netty.server.handler.message.BinaryMessageHandler;
import com.yonge.netty.server.processor.WaveformWriter;

@Component
public class CompareHandler implements BinaryMessageHandler {

	@Autowired
	private UserChannelContextService userChannelContextService;

	@Autowired
	private NettyChannelManager nettyChannelManager;

	/**
	 * @describe 采样率
	 */
	private float sampleRate = 44100;

	/**
	 * 每个采样大小（Bit）
	 */
	private int bitsPerSample = 16;

	/**
	 * 通道数
	 */
	private int channels = 1;

	/**
	 * @describe 采样大小
	 */
	private int bufferSize = 1024 * 4;

	private boolean signed = true;

	private boolean bigEndian = false;

	private AudioFormat audioFormat = new AudioFormat(sampleRate, bitsPerSample, channels, signed, bigEndian);

	private AudioFloatConverter converter = AudioFloatConverter.getConverter(audioFormat);

	private String tmpFileDir = "e:/soundRecords/";

	private SimpleDateFormat sdf = new SimpleDateFormat("yyMMddHHmmSS");

	@Override
	public String getAction() {
		return "SOUND_COMPARE";
	}

	@Override
	public boolean handler(String user, Channel channel, byte[] datas) {
		UserChannelContext channelContext = userChannelContextService.getChannelContext(channel);

		if (channelContext == null) {
			return false;
		}

		// 写录音文件
		WaveformWriter waveFileProcessor = channelContext.getWaveFileProcessor();
		if (waveFileProcessor == null) {
			File file = new File(tmpFileDir + user + "_" + sdf.format(new Date()) + ".wav");
			waveFileProcessor = new WaveformWriter(file.getAbsolutePath());
			channelContext.setWaveFileProcessor(waveFileProcessor);
		}
		waveFileProcessor.process(datas);

		datas = channelContext.skipHeader(datas);

		if (datas.length == 0) {
			return false;
		}

		channelContext.setChannelBufferBytes(ArrayUtil.mergeByte(channelContext.getChannelBufferBytes(), datas));

		int totalLength = channelContext.getChannelBufferBytes().length;

		while (totalLength >= bufferSize) {
			byte[] bufferData = ArrayUtil.extractByte(channelContext.getChannelBufferBytes(), 0, bufferSize - 1);

			if (bufferSize != totalLength) {
				channelContext.setChannelBufferBytes(ArrayUtil.extractByte(channelContext.getChannelBufferBytes(), bufferSize, totalLength - 1));
			} else {
				channelContext.setChannelBufferBytes(new byte[0]);
			}

			float[] sampleFloats = new float[bufferSize / 2];

			converter.toFloatArray(bufferData, sampleFloats);

			channelContext.handle(sampleFloats, audioFormat);

			MusicXmlBasicInfo musicXmlBasicInfo = channelContext.getMusicXmlBasicInfo(null);
			int sectionIndex = channelContext.getEvaluatingSectionIndex().get();

			// 评分
			int score = channelContext.evaluateForSection(sectionIndex, musicXmlBasicInfo.getSubjectId());
			if (score >= 0) {

				Map<String, Object> params = new HashMap<String, Object>();
				params.put("score", score);
				params.put("measureIndex", sectionIndex);

				WebSocketResponse<Map<String, Object>> resp = new WebSocketResponse<Map<String, Object>>("measureScore", params);

				nettyChannelManager.sendTextMessage(user, resp);
			}

			totalLength = channelContext.getChannelBufferBytes().length;
		}

		return true;
	}

}
