package com.ym.mec.biz.dal.dto;

import be.tarsos.dsp.AudioEvent;
import be.tarsos.dsp.SilenceDetector;
import be.tarsos.dsp.pitch.PitchDetectionHandler;
import be.tarsos.dsp.pitch.PitchDetectionResult;
import com.alibaba.fastjson.JSON;
import com.ym.mec.biz.dal.enums.DeviceTypeEnum;
import com.ym.mec.biz.dal.enums.HeardLevelEnum;
import com.ym.mec.biz.service.impl.SoundCompareHandler;
import io.swagger.annotations.ApiModelProperty;
import org.springframework.util.CollectionUtils;

import java.io.File;
import java.io.RandomAccessFile;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @Author Joburgess
 * @Date 2021/6/22 0022
 */
public class SoundCompareHelper implements PitchDetectionHandler {

    @ApiModelProperty(value = "设备类型")
    private DeviceTypeEnum deviceType;

    @ApiModelProperty(value = "操作编号")
    private String behaviorId;

    @ApiModelProperty(value = "伴奏编号")
    private Integer musicScoreId;

    @ApiModelProperty(value = "声部编号")
    private Integer subjectId;

    @ApiModelProperty(value = "评测记录编号")
    private Long recordId;

    @ApiModelProperty(value = "评测等级")
    private HeardLevelEnum heardLevel = HeardLevelEnum.ADVANCED;

    @ApiModelProperty(value = "小节开始时间")
    private double measureStartTime = 0;

    @ApiModelProperty
    private RandomAccessFile accessFile;

    private File file;

    @ApiModelProperty(value = "录音文件地址")
    private String recordFilePath;

    @ApiModelProperty(value = "播放速度")
    private int speed = 90;

    @ApiModelProperty(value = "xml中每个小节的音符详情")
    private Map<Integer, List<MusicPitchDetailDto>> measureXmlInfoMap = new HashMap<>();

    @ApiModelProperty(value = "xml中每小节开始时间、结束时间")
    private Map<Integer, MusicPitchDetailDto> measureEndTime = new HashMap<>();

//    @ApiModelProperty(value = "录音音频信息")
//    private List<MusicPitchDetailDto> recordMeasurePithInfo = new ArrayList<>();

    @ApiModelProperty(value = "小节分数记录")
    private Map<String, BigDecimal> userScoreMap = new HashMap<>();

    private Map<Integer, Map<String, Object>> userMeasureScoreMap = new HashMap<>();

    @ApiModelProperty(value = "xml中每个音符频率")
    private Map<Integer, Float> musicalNotePitchMap = new HashMap<>();

    @ApiModelProperty(value = "每个音符评测结果")
    private List<MusicalNotesPlayStatDto> musicalNotesPlayStats = new ArrayList<>();

    @ApiModelProperty(value = "偏移时间量，解决客户端录音播放不同步导致的声音留白")
    private int offsetTime = -1;

    /** xml中每个音符详情 */
    private List<MusicPitchDetailDto> musicXmlInfos;

    /** 第一个音符前有多少个字节 */
    private long firstMeasureStartBytes = 0;

    /** 文件的总字节数长度 */
    private long recordBytes = 0;

    private String clientId;

    private Integer detailId;

    private Integer examSongId;

    private String xmlUrl;

    /** 演奏音频数据中频率变更明显持续数量 */
    private int obviousChangeNum = 0;

    /** 演奏音频中但前小节所有频率数据 */
    private List<MusicPitchDetailDto> currPitchInfos = new ArrayList<>();
    private List<MusicPitchDetailDto> currTmpPitchInfos = new ArrayList<>();

    private List<MusicPitchDetailDto> recordMeasurePitchInfos = new ArrayList<>();

    /**
     * @describe 分贝检测器
     */
    public SilenceDetector silenceDetector = new SilenceDetector();

    /**
     * xml中每个音符详情 
     * @return
     */
    public List<MusicPitchDetailDto> getMusicXmlInfos() {
        return musicXmlInfos;
    }

    /**
     * xml中每个音符详情 
     * @return
     */
    public void setMusicXmlInfos(List<MusicPitchDetailDto> musicXmlInfos) {
        this.musicXmlInfos = musicXmlInfos;
    }

    /**
     * xml中每个音符对应的频率
     * @return
     */
    public Map<Integer, Float> getMusicalNotePitchMap() {
        return musicalNotePitchMap;
    }

    /**
     * xml中每个音符对应的频率
     * @return
     */
    public void setMusicalNotePitchMap(Map<Integer, Float> musicalNotePitchMap) {
        this.musicalNotePitchMap = musicalNotePitchMap;
    }

    public Long getRecordId() {
        return recordId;
    }

    public void setRecordId(Long recordId) {
        this.recordId = recordId;
    }

    public int getSpeed() {
        return speed;
    }

    public void setSpeed(int speed) {
        this.speed = speed;
    }

    public String getBehaviorId() {
        return behaviorId;
    }

    public void setBehaviorId(String behaviorId) {
        this.behaviorId = behaviorId;
    }

    public Integer getSubjectId() {
        return subjectId;
    }

    public void setSubjectId(Integer subjectId) {
        this.subjectId = subjectId;
    }

    public HeardLevelEnum getHeardLevel() {
        return heardLevel;
    }

    public void setHeardLevel(HeardLevelEnum heardLevel) {
        this.heardLevel = heardLevel;
    }

    public DeviceTypeEnum getDeviceType() {
        return deviceType;
    }

    public void setDeviceType(DeviceTypeEnum deviceType) {
        this.deviceType = deviceType;
    }

    public Integer getDetailId() {
        return detailId;
    }

    public void setDetailId(Integer detailId) {
        this.detailId = detailId;
    }

    public Integer getExamSongId() {
        return examSongId;
    }

    public void setExamSongId(Integer examSongId) {
        this.examSongId = examSongId;
    }

    public String getXmlUrl() {
        return xmlUrl;
    }

    public void setXmlUrl(String xmlUrl) {
        this.xmlUrl = xmlUrl;
    }

    public int getOffsetTime() {
        return offsetTime;
    }

    public void setOffsetTime(int offsetTime) {
        this.offsetTime = offsetTime;
    }

    public Integer getMusicScoreId() {
        return musicScoreId;
    }

    public void setMusicScoreId(Integer musicScoreId) {
        this.musicScoreId = musicScoreId;
    }

    /**
     * 录音文件
     * @return
     */
    public RandomAccessFile getAccessFile() {
        return accessFile;
    }

    public File getFile() {
        return file;
    }

    public void setFile(File file) {
        this.file = file;
    }

    public String getRecordFilePath() {
        return recordFilePath;
    }

    public void setRecordFilePath(String recordFilePath) {
        this.recordFilePath = recordFilePath;
    }

    /**
     * 录音文件
     * @return
     */
    public void setAccessFile(RandomAccessFile accessFile) {
        this.accessFile = accessFile;
    }

    public String getClientId() {
        return clientId;
    }

    public void setClientId(String clientId) {
        this.clientId = clientId;
    }

    public double getMeasureStartTime() {
        return measureStartTime;
    }

    public void setMeasureStartTime(double measureStartTime) {
        this.measureStartTime = measureStartTime;
    }

    /**
     * xml中每个小节的音符详情
     * @return
     */
    public Map<Integer, List<MusicPitchDetailDto>> getMeasureXmlInfoMap() {
        return measureXmlInfoMap;
    }

    /**
     * xml中每个小节的音符详情
     * @return
     */
    public void setMeasureXmlInfoMap(Map<Integer, List<MusicPitchDetailDto>> measureXmlInfoMap) {
        this.measureXmlInfoMap = measureXmlInfoMap;
    }

    /**
     * xml中每小节开始时间、结束时间
     * @return
     */
    public Map<Integer, MusicPitchDetailDto> getMeasureEndTime() {
        return measureEndTime;
    }

    /**
     * xml中每小节开始时间、结束时间
     * @param measureEndTime
     */
    public void setMeasureEndTime(Map<Integer, MusicPitchDetailDto> measureEndTime) {
        this.measureEndTime = measureEndTime;
    }

    /**
     * 每个音符的评测结果
     * @return
     */
    public List<MusicalNotesPlayStatDto> getMusicalNotesPlayStats() {
        return musicalNotesPlayStats;
    }

    /**
     * 每个音符的评测结果
     * @return
     */
    public void setMusicalNotesPlayStats(List<MusicalNotesPlayStatDto> musicalNotesPlayStats) {
        this.musicalNotesPlayStats = musicalNotesPlayStats;
    }

    /**
     * 小节累积得分（节奏、音准、完整度）
     * @return
     */
    public Map<String, BigDecimal> getUserScoreMap() {
        return userScoreMap;
    }

    /**
     * 小节累积得分（节奏、音准、完整度）
     * @return
     */
    public void setUserScoreMap(Map<String, BigDecimal> userScoreMap) {
        this.userScoreMap = userScoreMap;
    }

    /**
     * 第一个音符前有多少个字节
     * @return
     */
    public long getFirstMeasureStartBytes() {
        return firstMeasureStartBytes;
    }

    /**
     * 第一个音符前有多少个字节
     * @param firstMeasureStartBytes
     */
    public void setFirstMeasureStartBytes(long firstMeasureStartBytes) {
        this.firstMeasureStartBytes = firstMeasureStartBytes;
    }

    /**
     * 文件的总字节数长度
     * @return
     */
    public long getRecordBytes() {
        return recordBytes;
    }

    /**
     * 文件的总字节数长度
     * @param recordBytes
     */
    public void setRecordBytes(long recordBytes) {
        this.recordBytes = recordBytes;
    }

    /**
     * 小节得分结果
     * @return
     */
    public Map<Integer, Map<String, Object>> getUserMeasureScoreMap() {
        return userMeasureScoreMap;
    }

    /**
     * 小节得分结果
     * @return
     */
    public void setUserMeasureScoreMap(Map<Integer, Map<String, Object>> userMeasureScoreMap) {
        this.userMeasureScoreMap = userMeasureScoreMap;
    }

    public List<MusicPitchDetailDto> getRecordMeasurePitchInfos() {
        return recordMeasurePitchInfos;
    }

    public void setRecordMeasurePitchInfos(List<MusicPitchDetailDto> recordMeasurePitchInfos) {
        this.recordMeasurePitchInfos = recordMeasurePitchInfos;
    }

    @Override
    public void handlePitch(PitchDetectionResult pitchDetectionResult, AudioEvent audioEvent) {
        int timeStamp = (int) (measureStartTime + audioEvent.getTimeStamp()*1000);
        float pitch = pitchDetectionResult.getPitch();
        int decibel = (int) (100 - Math.abs(silenceDetector.currentSPL()));

        if(decibel <= SoundCompareHandler.soundCompareConfig.validDb){
            pitch = -1;
            decibel = 0;
        }

        SoundCompareHandler.LOGGER.info("时间：{}, 频率：{}, 分贝：{}", timeStamp, pitch, decibel);

        //上次的频率与本次的频率相差10hz或分贝相差10
        if(currPitchInfos.size()>0&&(Math.abs(currPitchInfos.get(currPitchInfos.size()-1).getFrequency()-pitch)>10||Math.abs(currPitchInfos.get(currPitchInfos.size()-1).getDecibel()-decibel)>10)){
            Double avgPitch = currPitchInfos.stream().skip(1).collect(Collectors.averagingDouble(MusicPitchDetailDto::getFrequency));
            Double avgDb = currPitchInfos.stream().skip(1).collect(Collectors.averagingDouble(MusicPitchDetailDto::getDecibel));

            MusicPitchDetailDto measureDetail = new MusicPitchDetailDto(currPitchInfos.get(0).getTimeStamp(), avgPitch.floatValue(), avgDb);
            measureDetail.setEndTimeStamp(timeStamp);
            measureDetail.setDuration(measureDetail.getEndTimeStamp()-measureDetail.getTimeStamp());

            if(measureDetail.getDuration()>25&&(avgPitch>0||avgDb>SoundCompareHandler.soundCompareConfig.validDb)){
                recordMeasurePitchInfos.add(measureDetail);
            }

            currPitchInfos.clear();

            //初始化偏移时间
            if(offsetTime == -1 && recordMeasurePitchInfos.size() == 1){
                offsetTime = recordMeasurePitchInfos.get(0).getTimeStamp();
                for (MusicPitchDetailDto musicXmlInfo : musicXmlInfos) {
                    if(!musicXmlInfo.getDontEvaluating()){
                        if(offsetTime > musicXmlInfo.getTimeStamp()) {
                            offsetTime = offsetTime - musicXmlInfo.getTimeStamp();
                        }
                        break;
                    }
                }
                musicXmlInfos.forEach(e->e.setTimeStamp(e.getTimeStamp()+offsetTime));//??????
                for (Map.Entry<Integer, MusicPitchDetailDto> musicPitchDetailDtoEntry : measureEndTime.entrySet()) {
                    musicPitchDetailDtoEntry.getValue().setTimeStamp(musicPitchDetailDtoEntry.getValue().getTimeStamp() + offsetTime);
                    musicPitchDetailDtoEntry.getValue().setEndTimeStamp(musicPitchDetailDtoEntry.getValue().getEndTimeStamp() + offsetTime);
                }
            }
        }
        currPitchInfos.add(new MusicPitchDetailDto(timeStamp, pitch, decibel));
    }
}
